<?php
/**
* @package		ConverseKit
* @copyright	Copyright (C) Stack Ideas Sdn Bhd. All rights reserved.
* @license		GNU/GPL, see LICENSE.php
* ConverseKit is free software. This version may have been modified pursuant
* to the GNU General Public License, and as distributed it includes or
* is derivative of works licensed under the GNU General Public License or
* other free or open source software licenses.
* See COPYRIGHT.php for copyright notices and details.
*/
defined('_JEXEC') or die('Unauthorized Access');

class ConversekitApi
{
	private $api = null;
	private $themeConfig = null;
	private $model = null;
	private $lib = null;

	// Response objects
	private $data = null;

	private $loadMoreForContacts = false;

	public function __construct(JRegistry $params)
	{
		// Ensure that EasySocial scripts are also rendered
		ES::language()->loadSite();

		$this->model = CK::model('Conversekit');
		$this->params = $params;
		$this->themeConfig = ES::themes()->getConfig();
		$this->lib = ES::conversations();
		$this->app = JFactory::getApplication();
		$this->input = $this->app->input;

		$this->data = new stdClass();
	}

	/**
	 * Allows user to exchange their username and password for a valid token
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	protected function getUserIdFromAuthKey($key = '')
	{
		if (!$key) {
			$key = $this->input->get('key', '', 'default');
		}

		if (!$key) {
			return false;
		}

		static $items = array();

		if (!isset($items[$key])) {
			$items[$key] = false;

			// Check if the provided key is valid
			$db = ES::db();
			$query = [
				'SELECT `user_id`',
				'FROM `#__social_users` AS a',
				'WHERE a.`auth` = ' . $db->Quote($key),
				'AND a.`state` = 1'
			];

			
			$db->setQuery($query);

			$userId = (int) $db->loadResult();

			if ($userId) {
				$items[$key] = $userId;
			}
		}

		return $items[$key];
	}

	/**
	 * SSE Event to retrieve ongoing states
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function heartbeat()
	{
		$this->isAuthKeyValid();

		$userId = $this->getUserIdFromAuthKey();
		$debug = $this->input->get('debug', false, 'bool');
		$isSSE = $this->params->get('single_polling', false);
		$interval = $this->params->get('monitor_interval');

		if ($isSSE && !$debug) {
			$this->setSSEHeader();
		}

		// Get the current time once
		$now = ES::date()->toSql();

		if ($isSSE) {
			while (true) {

				// Retrieve total number of unread items
				$totalUnread = $this->getTotalUnread($userId);
				echo $this->responseSSE('unread', $totalUnread);
				echo "\n\n";

				// Retrieve push notifications
				$pushData = $this->getPushData($userId, $now);

				if ($pushData) {
					echo $this->responseSSE('push', $pushData);
					echo "\n\n";
				}

				$this->setSSEEnding();

				if ($debug) {
					exit;
				}
			}
		}

		$data = array();
		$data['pushData'] = [];

		// Retrieve total number of unread items
		$data['totalUnread'] = $this->getTotalUnread($userId);

		// Retrieve push notifications
		$pushData = $this->getPushData($userId, $now);

		if ($pushData) {
			$data['pushData'] = $pushData;
		}

		$this->set('data', $data);

		return $this->response();
	}
	
	/**
	 * Initialize request from the app should go through here
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function initialize()
	{
		$this->isAuthKeyValid();

		$userId = (int) $this->getUserIdFromAuthKey();
		$viewer = $this->setCurrentUserInstance($userId);

		$conversations = $this->getConversations($userId);

		$contacts = $this->getContacts($userId);
		$hasLoadMoreForContacts = $this->hasLoadMoreForContacts();

		$settings = $this->getSettings($userId);

		$this->set('conversations', $conversations);
		$this->set('contacts', $contacts);
		$this->set('hasLoadMoreForContacts', $hasLoadMoreForContacts);
		$this->set('settings', $settings);

		return $this->response();
	}

	/**
	 * Determines if the provided attachment is allowed
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	protected function isAttachmentAllowed($name)
	{
		// Get supported extension
		$config = ES::config();
		$extensions = ES::makeString($config->get('conversations.attachments.types'), ',');
		$extensions = array_filter(explode(',', $extensions));
		
		$ext = JFile::getExt($name);

		if (!in_array($ext, $extensions)) {
			return false;
		}

		return true;
	}

	/**
	 * Determines if the auth key is valid
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	protected function isAuthKeyValid($key = '', $throwException = true)
	{
		$userId = $this->getUserIdFromAuthKey($key);

		if ($userId === false) {

			if ($throwException) {
				$this->reject('Invalid authentication key provided. Your ip will be logged');
			}

			return false;
		}

		return true;
	}

	/**
	 * Loads more conversations
	 *
	 * @since	2.0.3
	 * @access	public
	 */
	public function loadMoreConversations()
	{
		$this->isAuthKeyValid();

		$userId = (int) $this->getUserIdFromAuthKey();
		$viewer = $this->setCurrentUserInstance($userId);

		$conversations = $this->getConversations($userId, true);

		$this->set('conversations', $conversations);

		return $this->response();
	}

	/**
	 * Allows user to login in exchange for an authentication token
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function login()
	{
		$username = $this->input->get('username', '', 'default');
		$password = $this->input->get('password', '', 'default');

		$data = array('username' => $username, 'password' => $password);

		$state = $this->app->login($data);

		if ($state === false) {
			return $this->reject('Invalid username or password provided');
		}

		// Get the user's id based on the username.
		$model = ES::model('Users');
		$id = (int) $model->getUserId('username', $username);

		if (!$id) {
			return $this->reject('Unable to locate your user id, please try again');
		}

		// Generate a user authentication token
		$user = ES::user($id);

		if ($user->block) {
			return $this->reject('JERROR_NOLOGIN_BLOCKED');
		}

		$token = $user->getAuthToken(true);

		$user = ES::user($id);

		// Let easysocial conversekit app to know the user has logged into the app before.
		// The app will need to start storing push data in the respective table.
		$settings = $this->model->getSettings($id, $this->params);
		$settings['app'] = true;
		$settings = serialize($settings);
		
		$this->model->setSettings($id, $settings);

		$this->set('id', $id);
		$this->set('token', $token);
		$this->set('user', $user->toExportData($user));

		return $this->response();
	}

	/**
	 * Log the user out from the site
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function logout()
	{
		$this->app->logout();
	}

	/**
	 * Set the SSE Header
	 *
	 * @since	2.0.5
	 * @access	public
	 */
	private function setSSEHeader()
	{
		header("Cache-Control: no-cache");
		header("Content-Type: text/event-stream");
		header("X-Accel-Buffering: no");
	}

	/**
	 * Set the SSE's ending code
	 *
	 * @since	2.0.5
	 * @access	public
	 */
	private function setSSEEnding()
	{
		@ob_end_flush();
		@flush();

		// Some 3rd party plugin could block multiple request on same session. #4181 from ES
		// Refresh the session for each loop. #4181
		session_write_close();

		usleep(800000);
	}

	/**
	 * Method to monitor an active conversation
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function monitor()
	{
		$this->isAuthKeyValid();

		$userId = $this->getUserIdFromAuthKey();
		$ckSettings = $this->model->getSettings($userId, $this->params);
		$conversationId = $this->input->get('id', 0, 'int');
		$debug = $this->input->get('debug', false, 'bool');
		$markAsRead = $this->input->get('markAsRead', false, 'bool');
		$restApiKey = $this->input->get('key', '', 'default');

		// Ensure that the user can truly access the conversation
		$this->validateConversation($conversationId, $userId);

		$options = array();
		$options['single_polling'] = $this->params->get('single_polling', false);
		$options['monitor_interval'] = $this->params->get('monitor_interval');

		if ($options['single_polling'] && !$debug) {
			$this->setSSEHeader();
		}

		// We only start searching for new unread items from now.
		$timestamp = gmdate('Y-m-d H:i:s');

		if ($options['single_polling']) {
			// Keep a list of message ids that was previously sent so that we will not resend items that was sent before
			$exclusion = [];

			while (true) {
				// Check for new messages since now.
				$messages = $this->getNewMessages($conversationId, $userId, $timestamp, $restApiKey, $exclusion, $options);

				if ($messages) {
					foreach ($messages as $message) {
						$exclusion[] = (int) $message->id;
					}
				}

				echo $this->responseSSE('updates', $messages);
				echo "\n\n";

				// Determines if we should mark the conversation as read
				if ($markAsRead) {
					$this->lib->markAsRead($conversationId, $userId);
				}

				$this->setSSEEnding();

				if ($debug) {
					exit;
				}
			}
		}

		$messages = $this->getNewMessages($conversationId, $userId, $timestamp, $restApiKey, array(), $options);

		// Determines if we should mark the conversation as read
		if ($markAsRead) {
			$this->lib->markAsRead($conversationId, $userId);
		}

		$this->set('data', $messages);

		return $this->response();
	}

	/**
	 * Retrieves the total number of unread messages
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	private function getTotalUnread($userId)
	{
		$db = ES::db();

		$query = [
			'SELECT COUNT(1) FROM',
			'(',
			'SELECT COUNT(a.`conversation_id`) AS `total`',
			'FROM `#__social_conversations_message` AS a',
			'INNER JOIN `#__social_conversations_message_maps` AS b',
			'ON a.`id` = b.`message_id`',
			'WHERE b.`user_id`=' . $db->Quote($userId),
			'AND b.`state` = 1',
			'AND b.`isread` = 0',
			'AND a.`created_by` != ' . $db->Quote($userId),
			'GROUP BY a.`conversation_id`',
			') AS x'
		];

		$db->setQuery($query);
		$total = (int) $db->loadResult();

		return $total;
	}

	/**
	 * REST API to retrieve a single conversation
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function getConversation()
	{
		$this->isAuthKeyValid();

		$userId = (int) $this->getUserIdFromAuthKey();
		$id = $this->input->get('id', 0, 'int');

		// Globally set the user instance
		$user = $this->setCurrentUserInstance($userId);

		// Ensure that we have the conversation id
		if (!$id) {
			return $this->reject('Invalid conversation id provided');
		}

		// Ensure that the user can truly read the conversation
		$this->validateConversation($id, $userId);

		$restApiKey = $this->input->get('key', '', 'default');

		// Mark conversation as read since user is opening the conversation
		$this->lib->markAsRead($id, $userId);

		$data = $this->getConversationData($user, $id, $restApiKey);

		foreach ($data as $key => $value) {
			$this->set($key, $value);
		}

		return $this->response();
	}

	/**
	 * Retrieves metadata from a conversation so that the app can update the online status
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function getConversationInfo()
	{
		$this->isAuthKeyValid();

		$userId = (int) $this->getUserIdFromAuthKey();

		// Globally set the user instance
		$this->setCurrentUserInstance($userId);

		$id = $this->input->get('id', 0, 'int');

		$conversation = $this->validateConversation($id, $userId);

		// Determines if the user has access to this conversation
		$hasAccess = $conversation->hasAccess();

		if (!$hasAccess) {
			return $this->reject('You are not supposed to have access to this conversation');
		}

		$isOnline = false;
		$totalParticipants = 0;
		$lastSeen = '';

		if ($conversation->isSingle()) {
			$recipientId = $this->input->get('recipientId', 0, 'int');

			if (!$recipientId) {
				return $this->reject('Invalid contact id');
			}

			$user = ES::user($recipientId);

			$isOnline = $user->isOnline();
			$lastSeen = $user->getLastSeen();
		}     		

		if ($conversation->isMultiple()) {
			$totalParticipants = $this->model->getParticipants($userId, $conversation->id, array(), array('totalOnly' => true));		
		}     

		$this->set('isOnline', $isOnline);
		$this->set('lastSeen', $lastSeen);
		$this->set('totalParticipants', $totalParticipants);

		return $this->response();
	}


	/**
	 * Generate a standard conversation data
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	private function getConversationData($user, $conversationId, $restApiKey, $limit = 20, $start = 0)
	{
		$options = [
			'limit' => $limit,
			'start' => $start,
			'restApi' => $restApiKey
		];

		$conversation = ES::conversation($conversationId);

		$data = $this->model->getMessages($user->id, $conversationId, $options);

		// Only display the user permalink if that is not a group.
		$contact = null;

		if ($conversation->isSingle()) {

			// Since we know this is a single conversation recipient, we need to get the participant id
			$participants = $conversation->getParticipantsId();
			$participantId = $participants[0];
			$participant = ES::user($participantId);

			$contact = $participant->toExportData($user);
		}

		$result = new stdClass();
		$result->id = $conversationId;
		$result->title = $conversation->getTitle();
		$result->single = $conversation->isSingle();
		$result->totalMessages = (int) $data->totalmessages;
		$result->contact = $contact;
		$result->timestamp = $data->timestamp;
		$result->messages = $data->messages;

		// For group
		if (!$result->single) {
			// We only show the total participants in the group
			$result->totalParticipants = $this->model->getParticipants($user->id, $conversationId, array(), array('totalOnly' => true));
		}

		return $result;
	}

	/**
	 * REST API to retrieve list of conversations
	 *
	 * @since	2.0.0
	 * @access	private
	 */
	private function getConversations($userId, $loadMore = false)
	{
		$viewer = $this->setCurrentUserInstance($userId);

		// Default sorting and ordering
		$sorting = $this->input->get('sorting', $this->themeConfig->get('conversation_sorting', 'lastreplied'), 'default');
		$ordering = $this->input->get('ordering', $this->themeConfig->get('conversation_ordering', 'desc'), 'default');

		// Limit each requests to the hard limit specified here
		$limit = 70;

		$options = [
			'sorting' => $sorting,
			'ordering' => $ordering,
			'limit' => $limit
		];

		$model = ES::model('conversations');

		// Ensure that during loadmore, the limitstart does not exceed the total number of conversations that we have
		if ($loadMore) {
			$limitstart = $this->input->get('limitstart', 0, 'int');

			$total = $model->getTotalCount($userId, 'inbox');
			
			if ($total <= $limitstart) {
				return array();
			}
		}

		$result = $model->getConversations($userId, $options);
		$data = [];

		if ($result) {

			$today = JFactory::getDate('today');
			$apiKey = $this->input->get('key', '', 'default');

			foreach ($result as $row) {

				$item = $this->formatConversation($row, $apiKey, $viewer);

				$data[] = $item;
			}
		}

		return $data;
	}

	/**
	 * Format conversation to a standard dataset
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	private function formatConversation($conversation, $apiKey, SocialUser $viewer)
	{
		$item = $conversation->toExportData($viewer);
		$item->notification = $conversation->notification;
		$item->single = $conversation->isSingle();
		$item->online = false;

		// Use modified as the last replied date
		$item->lastModifiedDate = $conversation->getLastRepliedDate()->toSql();

		// Fix typo in ES core
		$item->recipients = $item->reciepients;

		// We do not need participants and recipients
		unset($item->participants);
		unset($item->reciepients);

		$date = $conversation->getLastRepliedDate();
		$item->displayTime = $date->format('Y-m-d H:i:s', false);

		if ($item->single) {
			$item->contact = $item->recipients[0];

			if (!isset($item->contact->isOnline)) {
				$user = ES::user($item->contact->id);

				$item->contact->isOnline = $user->isOnline();
			}
		}

		// During the initial retrieval of conversations, retrieve the list of messages in the conversation
		$item->conversationData = $this->getConversationData($viewer, $conversation->id, $apiKey);

		return $item;
	}

	/**
	 * Retrieve a list of contacts during initialization
	 *
	 * @since	2.0.0
	 * @access	private
	 */
	private function getContacts($userId, $loadMore = false)
	{
		// Filter contacts
		$filter = $this->input->get('filter', 'all', 'word');

		// Ordering
		$sorting = $this->input->get('sorting', $this->themeConfig->get('conversation_sorting'), 'default');
		$ordering = $this->input->get('ordering', $this->themeConfig->get('conversation_ordering'), 'default');

		$options = array(
			'sorting' => $sorting,
			'ordering' => $ordering
		);

		$config = ES::config();
		$model = ES::model('Friends');

		$options['state'] = SOCIAL_FRIENDS_STATE_FRIENDS;
		$options['idonly'] = true;
		$options['sort'] = 'lastseen';

		$results = array();
		$limit = 70;

		if ($config->get('conversations.nonfriend')) {
			$options['everyone'] = true;

			// Filter by online users only
			if ($filter == 'onlineusers') {
				$options['online'] = true;
			}
		}

		if ($filter == 'onlinefriends') {
			$options['online'] = true;
			$options['everyone'] = false;
		}

		if ($loadMore) {
			$limitstart = $this->input->get('limitstart', 0, 'int');

			$options['start'] = $limitstart;
		}	

		// We add 1 here to determine if there is anything more to load for the subsequent request
		$options['limit'] = $limit + 1;

		$ids = $this->model->getFriends($userId, $options);
		$total = $this->model->getTotalFriends();
		$hasLoadMore = ($total > $limit) ? true : false;

		// Update loadMoreForContacts as well so that loadMoreContacts() can get the state of the loadmore correctly
		$this->loadMoreForContacts = $hasLoadMore;

		// Remove the additional id if there is next pagination
		if ($hasLoadMore) {
			array_pop($ids);
		}

		$users = [];

		if ($ids) {
			ES::user($ids);

			$viewer = ES::user($userId);

			foreach ($ids as $uid) {
				$user = ES::user($uid);
				$data = $this->formatContact($user, $viewer);
			
				$users[] = $data;
			}
		}

		return $users;
	}

	/**
	 * Search contacts
	 *
	 * @since    2.0.4
	 * @access    public
	 */
	public function searchContacts()
	{
		$this->isAuthKeyValid();

		$userId = (int) $this->getUserIdFromAuthKey();
		$this->setCurrentUserInstance($userId);

		$config = ES::config();

		$query = $this->input->get('query', '', 'string');
		$type = $config->get('users.displayName');
		$nonFriend = $config->get('conversations.nonfriend');
		$privacy = $config->get('privacy.enabled');

		$options = array();
		$options['everyone'] = false;
		$options['privacy'] = 'post.message';

		if ($nonFriend) {
			$options['everyone'] = true;
		}

		if (!$privacy) {
			$options['privacy'] = false;
		}

		$results = $this->model->searchFriends($userId, $query, $type, $options);

		$contacts = array();

		if ($results) {
			$viewer = ES::user($userId);

			foreach ($results as $row) {
				$contacts[] = $this->formatContact($row, $viewer);
			}
		}

		$this->set('contacts', $contacts);

		return $this->response();
	}

	/**
	 * Format contact to a standard dataset
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	private function formatContact(SocialUser $contact, SocialUser $viewer)
	{
		$item = $contact->toExportData($viewer);
		$item->conversationId = 0;

		if ($viewer->id != $contact->id) {					
			$item->conversationId = $this->model->getConversationId($viewer->id, $contact->id);
		}

		return $item;
	}

	/**
	 * Determine if the contacts have next pagination
	 *
	 * @since    2.0.3
	 * @access    public
	 */
	private function hasLoadMoreForContacts()
	{
		return $this->loadMoreForContacts;
	}

	/**
	 * Loads more contacts
	 *
	 * @since    2.0.3
	 * @access    public
	 */
	public function loadMoreContacts()
	{
		$this->isAuthKeyValid();

		$userId = (int) $this->getUserIdFromAuthKey();
		$this->setCurrentUserInstance($userId);

		$contacts = $this->getContacts($userId, true);
		$hasLoadMore = $this->hasLoadMoreForContacts();

		$this->set('contacts', $contacts);
		$this->set('hasLoadMore', $hasLoadMore);

		return $this->response();
	}

	/**
	 * REST api to retrieve the necessary settings
	 *
	 * @since	2.0
	 * @access	public
	 */
	private function getSettings($userId)
	{
		$config = ES::config();
		$conversation = ES::conversation();
		$giphy = ES::giphy();
		$ckSettings = $this->model->getSettings($userId, $this->params);

		$settings = array(
			'conversation_sorting' => $config->get('conversations.sorting'),
			'conversation_giphy' => $giphy->isEnabledForConversations(),
			'conversation_location' => $config->get('conversations.location'),
			'conversation_attachment' => (int) $config->get('conversations.attachments.enabled'),
			'message_non_friends' => (int) $config->get('conversations.nonfriend'),
			'enter_to_submit' => (int) $ckSettings['submit'],
			'monitor_interval' => (int) $this->params->get('monitor_interval'),
			'single_polling' => (int) $this->params->get('single_polling', false)
		);

		return $settings;
	}

	/**
	 * Retrieve new messages from a conversation
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	private function getNewMessages($conversationId, $userId, $timestamp, $restApiKey, $exclusion = array(), $options = array())
	{

		$messages = [];
		$result = $this->model->getNewMessages($conversationId, $userId, $timestamp, $exclusion, $options);

		foreach ($result as $row) {
			$message = ES::table('ConversationMessage');
			$message->bind($row);
			
			$date = $message->getRepliedDate(false);

			// Standardize the formatting of new messages similar to the initial rendering of messages
			$message = $this->model->formatMessage($message, $restApiKey);

			// For displaying time we need to pass in correct format to be used in moment.js
			$message->displayTime = $date->format('Y-m-d H:i:s', false);

			unset($message->locations);

			$messages[] = $message;
		}

		return $messages;
	}

	/**
	 * Retrieves push notification data
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function getPushData($userId, $now)
	{
		// We always only process one at a time.
		$limit = 5;

		$db = ES::db();
		$query = [
			'SELECT * FROM `#__conversekit_push`',
			'WHERE `user_id`=' . $db->Quote($userId)
		];

		$created = 'AND `created` >= ' . $db->Quote($now);

		if (!$this->params->get('single_polling', false)) {
			$created = 'AND `created` >= ' . 'DATE_SUB(' . $db->Quote($now) . ', INTERVAL ' . $this->params->get('monitor_interval') . ' SECOND)';
		}

		$query[] = $created;
		$query[] = 'LIMIT 0,' . $limit;

		$db->setQuery($query);
		$items = $db->loadObjectList();

		if (!$items) {
			return false;
		}

		$result = [];
		$ids = [];

		$user = ES::user($userId);
		$restApiKey = $this->input->get('key', '', 'default');

		foreach ($items as $item) {
			$ids[] = $item->id;

			$obj = new stdClass();
			$obj->type = $item->type;
			$obj->created = $item->created;
			$obj->data = $item->data;

			$data = json_decode($item->data);

			if ($obj->type == 'sent' || $obj->type == 'received') {
				$message = ES::table('ConversationMessage');
				$message->load($data->messageId);

				$messages = array($message);
				$messages = $this->model->formatMessages($messages, $restApiKey);
				$data->message = $messages[0];

				$obj->data = json_encode($data);
			}

			$result[] = $obj;
		}

		// Once we retrieve it, we'll delete it immediately
		$query = 'DELETE FROM `#__conversekit_push` WHERE `id` IN(' . implode(',', $ids) . ')';

		$db->setQuery($query);
		$db->execute();

		return $result;
	}

	/**
	 * Load more messages
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function loadMore()
	{
		$this->isAuthKeyValid();

		$userId = (int) $this->getUserIdFromAuthKey();

		// Globally set the user instance
		$user = $this->setCurrentUserInstance($userId);

		$id = $this->input->get('id', 0, 'int');
		$timestamp = $this->input->get('timestamp', '', 'default');

		// Ensure that we have the conversation id
		if (!$id) {
			return $this->reject('Invalid conversation id provided');
		}

		$limit = $this->input->get('limit', 60, 'int');
		$restApiKey = $this->input->get('key', '', 'default');

		$options = [
			'limit' => $limit,
			'timestamp' => $timestamp,
			'restApiKey' => $restApiKey
		];

		// Limit should be incremented by 1 so that we know if there is anything else to load
		$options['limit'] += 1;

		$conversation = ES::conversation($id);

		$messages = $this->model->loadMore($userId, $id, $options);
		$total = count($messages);

		// If there are a total of items +1, it means load more should still be present
		$hasLoadmore = false;

		if ($total > $limit) {
			$hasLoadmore = true;
			unset($messages[$limit]);
		}

		// Reverse the array because the result should always be oldest first
		$messages = array_reverse($messages);

		$this->set('messages', $messages);

		return $this->response();
	}

	/**
	 * Previews a file
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function preview()
	{
		$this->isAuthKeyValid();

		$userId = (int) $this->getUserIdFromAuthKey();

		// Globally set the user instance
		$user = $this->setCurrentUserInstance($userId);
		$id = $this->input->get('id', 0, 'int');

		$file = ES::table('File');
		$file->load($id);

		if (!$file->id || !$id) {
			return $this->reject('Invalid file id');
		}

		// Load up the conversation message
		$message = ES::table('ConversationMessage');
		$message->load($file->uid);

		// Something went wrong with this discussion as it doesn't have participants
		if (!$message->id) {
			return $this->reject('Invalid file id');
		}

		$conversation = ES::conversation($message->conversation_id);

		// Something went wrong with this discussion as it doesn't have participants
		if (!$conversation->id) {
			return $this->reject('Invalid file id');
		}

		// Check if viewer is a participant
		if (!$conversation->isParticipant($user->id)) {
			return $this->reject('You are not allowed here');
		}

		$file->preview();
		exit;
	}

	/**
	 * Rejects a request
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function reject($message, $code = 500)
	{
		$this->set('code', $code);
		$this->set('message', $message);

		return $this->response($code);
	}

	/**
	 * Create a new conversation
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function create()
	{
		$this->isAuthKeyValid();

		$userId = (int) $this->getUserIdFromAuthKey();
		$postData = $this->input->post->getArray();

		if (!$postData['id']) {
			return $this->reject('Invalid receipient id');
		}

		$conversationLib = ES::conversation(); 
		$canCreate = $conversationLib->canCreate();

		if (!$canCreate) {
			return $this->reject('You are not allowed to create a new conversation.');
		}

		$isFriends = $conversationLib->isFriends();

		if (!$isFriends) {
			return $this->reject('You are not allowed to send messages to non-friends.');
		}

		// Globally set the user instance
		$user = $this->setCurrentUserInstance($userId);	

		$data = array();
		$data['uid'] = (int) $postData['id'];
		$data['message'] = $postData['message'];

		// Process attachments first
		$files = $_FILES;

		if ($files) {
			$data['upload-id'] = array();

			foreach ($files as $file) {
				$data['upload-id'][] = $this->processAttachment($file, $user);
			}
		}

		// Create the new conversation now
		$conversation = $this->lib->create($userId, $data);

		$conversation = ES::conversation($conversation->id);

		$apiKey = $this->input->get('key', '', 'default');
		$data = $this->formatConversation($conversation, $apiKey, $user);

		$message = $conversation->getLastMessage();
		$message = $this->lib->genMessage($message);

		// Format the attachments
		if ($message->attachments) {
			$restApiKey = $this->input->get('key', '', 'default');
			$message->attachments = CK::formatAttachments($message->attachments, $restApiKey);
		}

		$data->message = $message;

		$this->set('data', $data);

		return $this->response();
	}

	/**
	 * Delete a conversation
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function delete()
	{
		$this->isAuthKeyValid();

		$userId = (int) $this->getUserIdFromAuthKey();
		$conversationId = $this->input->get('conversationId', 0, 'int');

		$model = ES::model('Conversations');
		$state = $model->delete($conversationId, $userId);

		if (!$state) {
			return $this->reject('You are not allowed to delete the conversation.');
		}

		$this->set('id', $userId);
		$this->set('conversationId', $conversationId);

		return $this->response();
	}

	/**
	 * Allows user to download a file
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function download()
	{
		$this->isAuthKeyValid();

		$userId = (int) $this->getUserIdFromAuthKey();

		// Globally set the user instance
		$user = $this->setCurrentUserInstance($userId);
		$id = $this->input->get('id', 0, 'int');

		$file = ES::table('File');
		$file->load($id);

		if (!$file->id || !$id) {
			return $this->reject('Invalid file id');
		}

		// Load up the conversation message
		$message = ES::table('ConversationMessage');
		$message->load($file->uid);

		// Something went wrong with this discussion as it doesn't have participants
		if (!$message->id) {
			return $this->reject('Invalid file id');
		}

		$conversation = ES::conversation($message->conversation_id);

		// Something went wrong with this discussion as it doesn't have participants
		if (!$conversation->id) {
			return $this->reject('Invalid file id');
		}

		// Check if viewer is a participant
		if (!$conversation->isParticipant($user->id)) {
			return $this->reject('You are not allowed here');
		}

		$file->download();
		exit;
	}

	/**
	 * Get the giphies with the query given
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function getGiphy()
	{
		$this->isAuthKeyValid();

		$userId = (int) $this->getUserIdFromAuthKey();

		// Globally set the user instance
		$user = $this->setCurrentUserInstance($userId);

		// Get the search query
		$query = $this->input->post->get('query', '', 'string');

		$giphy = ES::giphy();
		$giphy->getData($query);

		$data = $giphy->toExportData();

		$this->set('data', $data);

		return $this->response();
	}

	/**
	 * Process uploaded files
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function processAttachment($file, $user)
	{
		$allowed = $this->isAttachmentAllowed($file['name']);

		if (!$allowed) {
			return $this->reject(JText::sprintf('You are trying to upload an invalid file, %1$s', $file['name']));
		}

		$attachment = $this->lib->addAttachment($user->id, $file);

		return $attachment->id;
	}

	/**
	 * Reply the conversation
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function reply()
	{
		$this->isAuthKeyValid();

		$userId = (int) $this->getUserIdFromAuthKey();

		// Globally set the user instance
		$user = $this->setCurrentUserInstance($userId);	

		$data = $this->input->post->getArray();

		$id = (int) $data['id']; 
		$conversation = ES::conversation($id);
		
		$conversation = $this->validateConversation($id, $userId);

		// Determines if the user has access to this conversation
		$hasAccess = $conversation->hasAccess();

		if (!$hasAccess) {
			return $this->reject('You are not supposed to have access to this conversation');
		}

		// Process attachments first
		$files = $_FILES;

		if ($files) {
			$data['upload-id'] = array();

			foreach ($files as $file) {
				$data['upload-id'][] = $this->processAttachment($file, $user);
			}
		}

		// Bind the post data
		$conversation->bind($data);
		$conversation->save();

		$conversation = ES::conversation($id);
		$data = $conversation->toExportData($user);
		$date = $conversation->getLastRepliedDate();

		$message = $conversation->getLastMessage();
		$message = $this->lib->genMessage($message);

		// Format the attachments
		if ($message->attachments) {
			$restApiKey = $this->input->get('key', '', 'default');
			$message->attachments = CK::formatAttachments($message->attachments, $restApiKey);
		}
		// We need to pass in the correct format into moment.js
		$data->displayTime = $date->format('Y-m-d H:i:s', false);
		$data->message = $message;

		$this->set('data', $data);
		
		return $this->response();
	}

	/**
	 * Responds to a request
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function response($code = 200)
	{
		http_response_code($code);

		header('Content-type: application/json; UTF-8');
		echo json_encode($this->data);

		exit;
	}

	/**
	 * Send a SSE response
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function responseSSE($event, $data)
	{
		ob_start();
?>
event: <?php echo $event;?>

<?php if (is_object($data) || is_array($data)) { ?>
data: <?php echo json_encode($data); ?>
<?php } else { ?>
data: <?php echo $data; ?>
<?php } ?>
<?php
		$contents = ob_get_contents();
		ob_end_clean();

		return $contents;
	}

	/**
	 * Because the user is not logged into Joomla, we need a way to set a global instance of the current user
	 * acrosss the site as the "logged in user"
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function setCurrentUserInstance($userId)
	{
		$user = ES::user($userId, false, true);

		return $user;
	}

	/**
	 * Allows caller to assign data which can be returned in the response
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	private function set($key, $value)
	{
		$this->data->$key = $value;
	}

	/**
	 * Determine if this is a valid conversation
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function validateConversation($id, $userId)
	{
		$conversation = ES::conversation($id);

		// If conversation id is invalid or not supplied, we need to throw some errors.
		if (!$id || !$conversation->id || !$conversation->isReadable($userId)) {
			return $this->reject('Invalid conversation id');
		}

		return $conversation;
	}

	/**
	 * Allows user to exchange their username and password for a valid token
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function verify()
	{
		$userId = $this->getUserIdFromAuthKey();

		if ($userId === false) {
			return $this->reject('Invalid authentication key provided. Your ip will be logged');
		}

		$this->set('id', $userId);

		return $this->response();
	}

	/**
	 * Allows user to exchange their username and password for a valid token
	 *
	 * @since	2.0.0
	 * @access	public
	 */
	public function version()
	{
		$this->set('version', '2.0.0');

		return $this->response();
	}
}