<?php
/**
* @package		ConverseKit
* @copyright	Copyright (C) Stack Ideas Sdn Bhd. All rights reserved.
* @license		GNU/GPL, see LICENSE.php
* ConverseKit is free software. This version may have been modified pursuant
* to the GNU General Public License, and as distributed it includes or
* is derivative of works licensed under the GNU General Public License or
* other free or open source software licenses.
* See COPYRIGHT.php for copyright notices and details.
*/
defined('_JEXEC') or die('Unauthorized Access');

class ConversekitEasySocial extends EasySocial
{
	private $api = null;
	private $themeConfig = null;

	// Conversekit model
	private $model = null;

	// EasySocial conversation library
	private $lib = null;

	public function __construct(JRegistry $params)
	{
		// Ensure that EasySocial scripts are also rendered
		ES::language()->loadSite();

		// Initialize the model
		$this->model = CK::model('Conversekit');

		parent::__construct();

		$this->params = $params;
		$this->themeConfig = ES::themes()->getConfig();
		$this->lib = ES::conversations();
	}

	/**
	 * Simulate the build users in EasySocial api
	 *
	 * @since   2.0
	 * @access  public
	 */
	public function buildUsers($total = 0, $users = array(), $prefix = 'contacts')
	{
		// heading object
		$data = new stdClass();
		$data->totalcontacts = $total;
		$data->{$prefix} = array();

		if (! $users) {
			return array();
		}

		if ($users) {
			foreach ($users as $user) {

				$obj = new stdClass();

				$obj->id = $user->id;
				$obj->name = $user->getName();
				$obj->avatar = $user->getAvatar();
				$obj->url = $user->getPermalink();
				$obj->isOnline = $user->isOnline();

				$lastseen = JText::_('PLG_CK_USERS_NEVER_LOGGED_IN');
				if ($user->getLastVisitDate() != '0000-00-00 00:00:00') {
					$lastseen = $user->getLastVisitDate('lapsed');
				}
				$obj->lastseen = $lastseen;

				$data->{$prefix}[] = $obj;
			}
		}

		return $data;
	}

	/**
	 * Retrieves a list of user id's that should be excluded from a conversation
	 * because they are already a participant of the conversation
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function getParticipantIds($conversationId, $userId)
	{
		$exclude = array();
		$data = $this->lib->getParticipants($userId, $conversationId);

		// Exclude participants
		foreach ($data->contacts as $participant) {
			$exclude[] = $participant->id;
		}

		return $exclude;
	}

	/**
	 * Since the conversations api doesn't support passing in exclusion list, we need to match it here ourselves
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function getAllowedParticipants($contacts, $exclusion)
	{
		if (!$exclusion || !$contacts) {
			return $contacts;
		}

		$allowed = array();

		foreach ($contacts as $contact) {
			if (in_array($contact->id, $exclusion)) {
				continue;
			}

			$allowed[] = $contact;
		}

		return $allowed;
	}

	/**
	 * Returns the formatted conversation title
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function getTitle($title)
	{
		$maxLength = 20;
		$length = ESJString::strlen($title);

		if ($length > $maxLength) {
			$title = ESJString::substr($title, 0, $maxLength) . JText::_('COM_EASYSOCIAL_ELLIPSES');
		}

		return $title;
	}

	/**
	 * Retrieves the settings
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function getSettings()
	{
		$settings = $this->model->getSettings($this->my->id, $this->params);

		return $settings;
	}

	/**
	 * Renders the settings page
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function renderSettings()
	{
		$this->validateUser();

		$settings = $this->getSettings();
		$output = CK::getTemplateContents('settings', $settings);

		echo $output;
		exit;
	}

	/**
	 * Prevents malicious spoofing of user id's.
	 *
	 * @since	1.0.4
	 * @access	public
	 */
	private function validateUser()
	{
		if ($this->my->guest) {
			return $this->showError(JText::_('PLG_CK_NOT_ALLOWED'));
		}

		return true;
	}

	/**
	 * Retrieves a list of conversations on the site for the current viewer
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function renderConversations($isloadmore = false)
	{
		$this->validateUser();

		ES::checkToken();
		ES::requireLogin();

		$userId = $this->my->id;
		$sorting = $this->input->get('sorting', $this->themeConfig->get('conversation_sorting', 'lastreplied'), 'default');
		$ordering = $this->input->get('ordering', $this->themeConfig->get('conversation_ordering', 'desc'), 'default');

		$limit = $this->input->get('limit', $this->params->get('conversation_limit'), 'default');
		$start = $this->input->get('limitstart', 0, 'int');

		$options = [
			'sorting' => $sorting,
			'ordering' => $ordering,
			'limit' => $limit
		];

		$model = ES::model('conversations');
		$data = $model->getConversations($userId, $options);
		$total = $model->getTotal();

		if (!$data) {
			$data = new stdClass();
			$data->empty = true;

			return $this->output($data);
		}

		if ($data) {
			foreach ($data as &$item) {
				$item->participants = $item->getParticipants();
				$item->participant = $item->participants[0];
			}
		}

		$data = [
			'conversations' => $data, 
			'showLoadmore' => true, 
			'userId' => $userId
		];

		if ($isloadmore) {
			$data['showLoadmore'] = false;
		}

		if ($total <= ($start + $limit)) {
			$data['showLoadmore'] = false;
		}

		$contents = CK::getTemplateContents('conversations', $data);

		if ($isloadmore) {

			$doNext = ($total > ($start + $limit)) ? 1 : 0;

			return array($doNext, $contents);
		}


		// $contents = CK::getTemplateContents('conversations', $data);

		$data = new stdClass();
		$data->contents = $contents;
		$data->timestamp = ES::date()->toSql();

		return $this->output($data);
		exit;
	}

	/**
	 * Renders the last message for each of the conversations
	 *
	 * @since	2.0
	 * @access	public
	 */
	public function getLastMessage(SocialConversation $conversation)
	{
		$message = ES::string()->parseBBCode($conversation->getLastMessage()->getIntro(35), array('escape' => false, 'emoticons' => true, 'links' => false));

		if ($conversation->getLastMessageType() == 'giphy') {
			if ($conversation->getLastMessage()->created_by == $this->my->id) {
				$message = '<i class="fa fa-share"></i>&nbsp;' . JText::_('COM_ES_CONVERSATIONS_GIPHY_YOU_SENT_GIF');
			} else {
				$message = '&nbsp;' . JText::sprintf('COM_ES_CONVERSATIONS_GIPHY_OTHER_SENT_GIF', ES::user($conversation->getLastMessage()->created_by)->getName());
			}
		}

		return $message;
	}

	/**
	 * Marks a conversation as read
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function markAsRead()
	{
		ES::checkToken();
		ES::requireLogin();

		$userId = $this->my->id;
		$ids = $this->input->get('id', array(), 'array');

		foreach ($ids as $id) {
			$id = (int) $id;

			$this->lib->markAsRead($id, $userId);
		}

		$data = new stdClass();
		$data->message = JText::_('PLG_CK_READ_SUCCESS');

		$this->output($data);
	}

	/**
	 * Archives conversations
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function archive()
	{
		ES::checkToken();
		ES::requireLogin();

		$userId = $this->my->id;
		$ids = $this->input->get('id', array(), 'array');

		$output = array();

		foreach ($ids as $id) {
			$id = (int) $id;

			$output[] = $this->lib->archive($userId, $id);
		}

		$this->output($output);
	}

	/**
	 * Allows caller to save settings
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function saveSettings()
	{
		ES::checkToken();
		ES::requireLogin();

		$userId = $this->my->id;
		$sound = $this->input->get('sound', 0, 'int');
		$submit = $this->input->get('submit', 0, 'int');
		$darkMode = $this->input->get('darkmode', 0, 'int');

		$current = $this->model->getSettings($this->my->id, $this->params);


		$data = array(
				"sound" => $sound,
				"submit" => $submit,
				"darkmode" => $darkMode,
				"app" => isset($current['app']) ? $current['app'] : false
		);

		$data = serialize($data);

		$data = $this->model->setSettings($userId, $data);
		$data = $this->output($data);

		return $data;
	}

	/**
	 * Function to retrieve new messages from any conversations
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function getNotificationConversations()
	{
		ES::checkToken();
		ES::requireLogin();

		$userId = $this->my->id;

		$obj = new stdClass();
		$currentModif = null;

		$timestamp = $this->input->getString('timestamp');
		$data = $this->model->getNotification($userId);

		$counter = count($data);
		$conversationId = 0;
		$notification = 0;

		$messages = array();

		if ($counter > 0) {
			//get the last items date
			$currentModif = $data[$counter-1]->lastmodif;

			foreach ($data as $row) {
				$messages[$row->conversation_id][] = $row;
			}
		}

		// lets prepare the data
		$data = array();
		if ($messages) {
			foreach ($messages as $conversationId => $items) {

				$obj = new stdClass();

				$obj->conversationId = $conversationId;
				$obj->count = count($items);

				$lastMessage = $items[$obj->count - 1];

				$message = ES::table('ConversationMessage');
				$message->bind($lastMessage);

				$formattedMessage = ES::string()->parseBBCode($message->getIntro(150), array('escape' => false, 'emoticons' => true, 'links' => false));
				$obj->preview = $formattedMessage;
				$obj->dateLapse = $message->getRepliedDate();

				$data[] = $obj;
			}
		}

		if (gettype($currentModif) == 'object') {
			$currentModif = $currentModif->toSql(false);
		} else {
			$currentModif = $currentModif;
		}

		$obj = new stdClass();
		$obj->lastmodif = $currentModif;
		$obj->total = count($messages);
		$obj->data = $data;

		return $this->output($obj);
	}


	/**
	 * When the script determines there are new messages, we need to retrieve these data
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function getNotificationMessage()
	{
		ES::checkToken();
		ES::requireLogin();

		$userId = $this->my->id;

		$obj = new stdClass();
		$currentModif = null;

		$id = $this->input->getInt('id');
		$timestamp = $this->input->getString('timestamp');

		// get data
		$data = $this->model->getNotification($userId, $timestamp, $id);

		$counter = count($data);
		$conversationId = 0;
		$notification = 0;

		if ($counter > 0) {
			//get the last items date
			$currentModif = $data[$counter-1]->lastmodif;

			foreach ($data as $row) {

				$conversationId = $row->conversation_id;
				$notification = $row->newMsg;

				$message = ES::table('ConversationMessage');
				$message->bind($row);

				// format the messages so that the template can understand.
				$obj->messages[] = $this->lib->genMessage($message);
			}
		}

		$obj->contents = CK::getTemplateContents('message', array('data' => $obj));
		$obj->conversationId = $conversationId;
		$obj->notification = $notification;

		if (gettype($currentModif) == 'object') {
			$currentModif = $currentModif->toSql(false);
		} else {
			$currentModif = $currentModif;
		}

		$obj->lastmodif = $currentModif;

		// now we mark this converstation as read
		$this->lib->markAsRead($conversationId, $userId);

		return $this->output($obj);
	}

	/**
	 * load the next page for contacts
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function loadMoreContacts()
	{
		$this->validateUser();

		ES::checkToken();
		ES::requireLogin();

		$search = $this->input->get('search', null, 'string');
		$start = $this->input->get('limitstart', 0, 'int');
		$limit = $this->input->get('limit', $this->params->get('contact_limit'), 'default');
		$filter = $this->input->get('filter', '', 'default');

		$contents = "";
		$nextlimit = $start + $limit;
		$doNext = 0;

		if ($filter == 'participants') {
			// load participants
			$data = $this->renderParticipants(true);
			$doNext = $data[0];
			$contents = $data[1];
		} else if (is_null($search)) {

			// load contacts
			$data = $this->renderContacts(true);
			$doNext = $data[0];
			$contents = $data[1];

		} else {
			// load search
			$data = $this->searchContacts(true);
			$doNext = $data[0];
			$contents = $data[1];
		}

		$data = new stdClass();
		$data->nextlimit = $nextlimit;

		if (!$doNext) {
			$data->nextlimit = -1;
		}

		$data->contents = $contents;

		return $this->output($data);
	}


	/**
	 * load the next page for conversations
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function loadMoreConversations()
	{
		$this->validateUser();

		ES::checkToken();
		ES::requireLogin();

		$start = $this->input->get('limitstart', 0, 'int');
		$limit = $this->input->get('limit', $this->params->get('conversation_limit'), 'default');

		$contents = "";
		$nextlimit = $start + $limit;
		$doNext = 0;

		// load search
		$data = $this->renderConversations(true);
		$doNext = $data[0];
		$contents = $data[1];

		$data = new stdClass();
		$data->nextlimit = $nextlimit;

		if (!$doNext) {
			$data->nextlimit = -1;
		}

		$data->contents = $contents;

		return $this->output($data);
	}


	/**
	 * load the next page for messages
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function loadMoreMessages()
	{
		$this->validateUser();

		ES::checkToken();
		ES::requireLogin();

		$start = $this->input->get('limitstart', 0, 'int');
		$limit = $this->input->get('limit', $this->params->get('conversation_limit'), 'default');
		$id = $this->input->get('id', 0, 'int');


		$contents = "";
		$nextlimit = $start + $limit;
		$doNext = 0;

		// load search
		$data = $this->renderMessage($id, true);
		$doNext = $data[0];
		$contents = $data[1];

		$data = new stdClass();
		$data->nextlimit = $nextlimit;

		if (!$doNext) {
			$data->nextlimit = -1;
		}

		$data->contents = $contents;

		return $this->output($data);
	}

	/**
	 * Retrieves a list of contacts for a particular user
	 *
	 * @since	1.0.7
	 * @access	public
	 */
	public function renderContacts($isloadmore = false)
	{
		$this->validateUser();

		ES::checkToken();
		ES::requireLogin();

		$userId = $this->my->id;
		$filter = $this->input->get('filter', 'all', 'word');

		// Ordering
		$sorting = $this->input->get('sorting', $this->themeConfig->get('conversation_sorting'), 'default');
		$ordering = $this->input->get('ordering', $this->themeConfig->get('conversation_ordering'), 'default');

		// Pagination
		$limit = $this->input->get('limit', $this->params->get('contact_limit'), 'default');
		$start = $this->input->get('limitstart', 0, 'int');

		$options = array(
			'sorting' => $sorting,
			'ordering' => $ordering,
			'limit' => $limit + 1,
			'start' => $start
		);

		// This determines if viewer wants to add a participant to an existing conversation
		$addParticipant = $this->input->get('addParticipant', false, 'bool');
		$conversationId = $this->input->get('conversationId', 0, 'int');
		$excludeIds = array();

		// Always exclude the current user since there is no way for users to chat with themselves
		$excludeIds[] = $this->my->id;
		$excludeCount = 0;

		if ($addParticipant) {

			$participants = $this->getParticipantIds($conversationId, $this->my->id);

			// we need to minus off our self
			$excludeCount = count($participants) - 1;

			$excludeIds = array_merge($excludeIds, $participants);
		}

		$config = ES::config();
		$model = ES::model('Friends');

		$options['state'] = SOCIAL_FRIENDS_STATE_FRIENDS;
		$options['idonly'] = true;
		$options['sort'] = 'lastseen';

		$results = array();
		$total = 0;

		if ($this->config->get('conversations.nonfriend')) {
			$options['everyone'] = true;

			// Filter by online users only
			if ($filter == 'onlineusers') {
				$options['online'] = true;
			}
		}

		if ($filter == 'onlinefriends') {
			$options['online'] = true;
			$options['everyone'] = false;
		}

		$options['exclude'] = $excludeIds;
		$ids = $this->model->getFriends($userId, $options);
		$total = $this->model->getTotalFriends();
		$hasLoadMore = ($total > $limit) ? true : false;

		// Remove the additional id and minus 1 back to get the correct total number if there is next pagination
		if ($hasLoadMore) {
			array_pop($ids);

			$total = $total - 1;
		}

		if ($ids) {
			// preload users
			ES::user($ids);

			foreach ($ids as $uid) {
				$results[] = ES::user($uid);
			}
		}

		$data = $this->buildUsers($total, $results);

		// Get allowed participants
		if ($filter == 'onlinefriends' && $data) {
			$data->contacts = $this->getAllowedParticipants($data->contacts, $excludeIds);
		}

		if ($excludeCount) {
			$total = $total - $excludeCount;
		}

		// Only add users that are
		$args = array('contacts' => array(), 'addParticipant' => $addParticipant, 'showLoadmore' => true);

		if ($data) {
			$args['contacts'] = $data->contacts;
		} else {
			$args['showLoadmore'] = false;
		}

		if ($isloadmore) {
			$args['showLoadmore'] = false;
		}

		if (!$hasLoadMore) {
			$args['showLoadmore'] = false;
		}

		$contents = CK::getTemplateContents('contacts', $args);

		if ($isloadmore) {
			$doNext = $hasLoadMore;

			return array($doNext, $contents);
		}

		echo $contents;
		exit;
	}

	/**
	 * Similar to @renderMessage but this method takes a user id as an argument.
	 * It will then try to find a conversation id.
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function renderMessageWithUserId()
	{
		ES::checkToken();
		ES::requireLogin();

		$recipientId = $this->input->get('userId', 0, 'int');
		$userId = $this->my->id;

		if (!$recipientId) {
			$this->showError('Invalid recipient id provided.');
		}

		$limit = $this->input->get('limit', $this->params->get('conversation_limit'), 'default');
		$start = $this->input->get('limitstart', 0, 'int');

		$options = array(
						'limit' => $limit,
						'start' => $start
					);

		// Check to see if the user already has an existing conversation with the recipient
		$table = $this->model->getSingleConversation($userId, $recipientId);

		// When conversation already exists, we just render the message
		if ($table->id) {

			$conversation = ES::conversation($table);

			// Only proceed with rendering existing conversation if it has a message
			if ($conversation->getMessages()) {
				return $this->renderMessage($table->id);
			}
		}

		$recipient = ES::user($recipientId);

		$data = new stdClass();
		$data->conversationTitle = $this->getTitle($recipient->getName());
		$data->userOnlineStatus = $recipient->isOnline();
		$data->avatar = $recipient->getAvatar();
		$data->multiple = false;
		$data->isParticipant = true;
		$data->target = null;
		$data->userPermalink = 'javascript:void(0);';

		// Retrieve the one to one conversation user permalink
		if (!$data->multiple) {
			$participant = ES::user($recipientId);

			$data->userPermalink = $participant->getPermalink();
			$data->target = $participant;
		}

		// Get target user's last seen date
		$lastSeen = $recipient->getLastVisitDate();

		if ($lastSeen != '0000-00-00 00:00:00') {
			$lastSeen = $recipient->getLastVisitDate('lapsed');
		}

		$data->conversationLastActive = $lastSeen;

		$args = array('data' => $data);

		$output = new stdClass();

		$output->header = CK::getTemplateContents('message.header', $args);
		$output->contents = '';
		$output->isNew = true;
		$output->isGroup = false;

		// Since this is a new conversation, reply form should always be visible
		$output->canReply = true;
		$output->timestamp = ES::date()->toSql();

		$data = $this->output($output);

		return $data;
	}

	/**
	 * Retrieves messages in a conversation
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function renderMessage($id = null, $isloadmore = false)
	{
		$this->validateUser();

		ES::checkToken();
		ES::requireLogin();

		if ($id) {
			$id = (int) $id;
		} else {
			$id = $this->input->get('id', 0, 'int');
		}

		$userId = $this->my->id;

		if (!$id) {
			$this->showError(JText::_('PLG_CK_INVALID_ID'));
		}

		$limit = $this->input->get('limit', $this->params->get('conversation_limit'), 'default');
		$start = $this->input->get('limitstart', 0, 'int');

		$options = [
						'limit' => $limit,
						'start' => $start
		];

		$conversation = ES::conversation($id);

		// Do not allow unauthorized access to conversations.
		if (!$conversation->isReadable($userId)) {
			die('Disallowed');
		}

		$this->lib->markAsRead($id, $userId);

		$data = $this->model->getMessages($userId, $id, $options);
		$total = $data->totalmessages;

		$data->header = $this->lib->getParticipants($userId, $id);
		$data->conversationTitle = $this->getTitle($conversation->getTitle($id));
		$data->conversationAvatar = $conversation->getAvatar(false);
		$data->type = $conversation->type;
		$data->multiple = $data->type == 1 ? false : true;
		$data->avatar = $data->conversationAvatar[0];
		$data->target = null;
		$participant = null;

		$data->userPermalink = 'javascript:void(0);';

		// Retrieve the one to one conversation user permalink
		if (!$data->multiple) {
			$participant = $this->model->getParticipants($userId, $id, array($userId), $options);

			$data->target = $participant[0];

			if ($participant) {
				$data->userPermalink = ES::user($participant[0]->id)->getPermalink();
			}
		}


		if ($data->multiple) {
			$data->avatar = rtrim(JURI::root(), '/') . '/plugins/system/conversekit/assets/images/people.png';
		}

		$lastMessage = $conversation->getLastMessage();

		// Get logged in user's details
		$data->loggedInUserName = $this->my->getName();
		$data->loggedInUserAvatar = $this->my->getAvatar();

		// Check user online status if there's only two or lesser participants
		if ($data->header->totalcontacts <= 2) {

			for ($i = 0; $i < $data->header->totalcontacts; $i++) {
				if ($data->header->contacts[$i]->id != $userId) {

					$contactUser = ES::user($data->header->contacts[$i]->id);

					$lastSeen = $contactUser->getLastVisitDate();

					if ($lastSeen != '0000-00-00 00:00:00') {
						$lastSeen = $contactUser->getLastVisitDate('lapsed');
					}

					$data->conversationLastActive = $lastSeen;
					$data->userOnlineStatus = $contactUser->isOnline();
				}
			}
		} else {
			$data->userOnlineStatus = ES::user($lastMessage->created_by)->isOnline();
			$data->conversationLastActive = $lastMessage->getRepliedDate();
		}

		// Check participant status
		$data->isParticipant = $this->isParticipant($conversation->id, $this->my->id);


		$args = array('data' => $data, 'showLoadmore' => true);

		if ($isloadmore) {
			$args['showLoadmore'] = false;
		}

		if ($total <= ($start + $limit)) {
			$args['showLoadmore'] = false;
		}

		$contents = CK::getTemplateContents('message', $args);

		if ($isloadmore) {
			$doNext = ($total > ($start + $limit)) ? 1 : 0;
			return array($doNext, $contents);
		}

		$output = new stdClass();
		$output->id = $id;
		$output->header = CK::getTemplateContents('message.header', $args);
		$output->contents = $contents;
		$output->canReply = $data->isParticipant;
		$output->isNew = false;
		$output->isGroup = $data->multiple;
		$output->timestamp = ES::date()->toSql();

		$data = $this->output($output);

		return $data;
	}

	/**
	 * Determines if the user is a participant of a conversation
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	protected function isParticipant($conversationId, $userId)
	{
		$result = $this->lib->getParticipants($userId, $conversationId);

		// No participants in this conversation
		if (!$result->contacts || $result->totalcontacts <= 0) {
			return false;
		}

		$participants = array();

		foreach ($result->contacts as $participant) {
			$participants[] = (int) $participant->id;
		}

		if (in_array($userId, $participants)) {
			return true;
		}

		return false;
	}

	/**
	 * Allows user to store attachments
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function uploadAttachment()
	{
		ES::checkToken();
		ES::requireLogin();

		$file = $this->input->files->get('file');

		// Get supported extension
		$config = ES::config();
		$extensions = ES::makeString($config->get('conversations.attachments.types'), ',');
		$extensions = array_filter(explode(',', $extensions));

		$ext = JFile::getExt($file['name']);

		if (!in_array($ext, $extensions)) {
			return $this->showError(JText::_('PLG_CK_INVALID_ATTACHMENT'));
		}

		$size = round($file['size'] / 1024);
		$type = $file['type'];

		$attachment = $this->lib->addAttachment($this->my->id, $file);

		if (isset($attachment->code) && $attachment->code == 403) {
			return $this->showError($attachment);
		}

		return $this->output($attachment);
	}

	/**
	 * Deletes an attachment from a message
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function deleteAttachment()
	{
		ES::checkToken();
		ES::requireLogin();

		$id = $this->input->get('fileId');

		$data = $this->lib->deleteAttachment($this->my->id, $id);
		$data = $this->output($data);

		return $data;
	}

	/**
	 * Allows caller to search for contacts
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function searchContacts($isloadmore = false)
	{
		ES::checkToken();
		ES::requireLogin();

		$config = ES::config();
		$search = $this->input->get('search', '', 'string');
		$limit = $this->input->get('limit', $this->params->get('contact_limit'), 'default');
		$start = $this->input->get('limitstart', 0, 'int');
		$addParticipant = $this->input->get('addParticipant', false, 'bool');
		$conversationId = $this->input->get('conversationId', 0, 'int');

		$searchOptions = array('limit' => $limit, 'limitstart' => $start);


		$exclusion = array();

		if ($addParticipant) {
			$exclusion = $this->getParticipantIds($conversationId, $this->my->id);
		}


		if ($config->get('conversations.nonfriend')) {
			$searchOptions['everyone'] = true;
		}

		$searchOptions['privacy'] = 'profiles.post.message';

		// Determine what type of string we should search for.
		$type = $config->get('users.displayName');

		$userId = $this->my->id;

		$model = ES::model('Friends');
		$results = $model->search($userId, $search, $type, $searchOptions);
		$total = $model->getTotal();

		$data = $this->buildUsers($total, $results);

		// Get allowed participants
		if (isset($data->contacts)) {

			// rebuild the last seen
			$users = array();
			foreach ($data->contacts as $item) {
				$user = ES::user($item->id);

				$lastSeen = $user->getLastVisitDate();
				if ($lastSeen == '0000-00-00 00:00:00') {
					$item->lastseen = JText::_('PLG_CK_USERS_NEVER_LOGGED_IN');
				}
				$users[] = $item;
			}

			$data->contacts = $users;
			$data->contacts = $this->getAllowedParticipants($data->contacts, $exclusion);
		}

		$args = array('contacts' => array(), 'addParticipant' => $addParticipant, 'showLoadmore' => true);

		if ($data) {
			$args['contacts'] = isset($data->contacts) ? $data->contacts : array();
		} else {
			$args['showLoadmore'] = false;
		}

		if ($isloadmore) {
			$args['showLoadmore'] = false;
		}

		if ($total <= ($start + $limit)) {
			$args['showLoadmore'] = false;
		}

		$contents = CK::getTemplateContents('contacts', $args);

		if ($isloadmore) {

			$doNext = ($total > ($start + $limit)) ? 1 : 0;

			return array($doNext, $contents);
		}


		echo $contents;
		exit;
	}

	/**
	 * Renders a list of participants in a conversation
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function renderParticipants($isloadmore = false)
	{
		$this->validateUser();

		ES::checkToken();
		ES::requireLogin();

		$id = $this->input->get('id', 0, 'int');
		$userId = $this->my->id;
		$limit = $this->input->get('limit', $this->params->get('contact_limit'), 'default');
		$start = $this->input->get('limitstart', 0, 'int');

		$options = array(
			'limit' => $limit,
			'start' => $start
		);


		if (!$id) {
			$this->showError(JText::_('PLG_CK_INVALID_ID'));
		}

		$data = new stdClass();
		$conversation = ES::conversation($id);

		$participants = array();

		if ($conversation->id) {
			//$participants = $this->lib->getParticipants($userId, $id);

			$participants = $this->model->getParticipants($userId, $id, array(), $options);

			$options['totalOnly'] = true;
			$total = $this->model->getParticipants($userId, $id, array(), $options);

			$data = $this->buildUsers($total, $participants);
			$participants = $data->contacts;
		}

		$args = array('contacts' => array(), 'addParticipant' => false, 'showLoadmore' => true);

		if ($participants) {
			$args['contacts'] = $participants;
		} else {
			$args['showLoadmore'] = false;
		}

		if ($isloadmore) {
			$args['showLoadmore'] = false;
		}

		if ($total <= ($start + $limit)) {
			$args['showLoadmore'] = false;
		}

		$contents = CK::getTemplateContents('contacts', $args);

		if ($isloadmore) {

			$doNext = ($total > ($start + $limit)) ? 1 : 0;

			return array($doNext, $contents);
		}


		$contents = CK::getTemplateContents('contacts', $args);

		echo $contents;
		exit;
	}

	/**
	 * Allows remote caller to add participants into an existing conversation
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function addParticipants()
	{
		ES::checkToken();
		ES::requireLogin();

		$conversationId = $this->input->get('conversationId', 0, 'int');
		$participantId = $this->input->get('participantId', 0, 'int');

		if (!$conversationId || !$participantId) {
			$this->showError(JText::_('PLG_CK_INVALID_ID'));
		}

		$data = $this->lib->addParticipant($this->my->id, $conversationId, $participantId);
		$data = $this->output($data);
		return $data;
	}

	/**
	 * Allows caller to leave a conversation
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function leave()
	{
		ES::checkToken();
		ES::requireLogin();

		$conversationId = $this->input->get('conversationId', 0, 'int');

		if (!$conversationId) {
			$this->showError(JText::_('PLG_CK_INVALID_ID'));
		}

		$data = $this->lib->deleteParticipant($conversationId, $this->my->id);

		$data = $this->output($data);
		return $data;
	}

	/**
	 * Allows caller to delete conversations from the site
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function deleteConversations()
	{
		ES::checkToken();
		ES::requireLogin();

		$ids = $this->input->get('id', array(), 'array');

		if (!$ids) {
			$this->showError(JText::_('PLG_CK_INVALID_ID'));
		}

		foreach ($ids as $id) {
			$id = (int) $id;

			$this->lib->delete($this->my->id, $id);
		}

		$response = new stdClass();
		$response->message = JText::_('PLG_CK_DELETED');

		return $this->output($response);
	}

	/**
	 * Search for giphy based on the query given
	 *
	 * @since	2.0
	 * @access	public
	 */
	public function searchGiphy()
	{
		$giphy = ES::giphy();
		$data = $this->input->getArray('post');

		// Search and get the data
		$data = ES::giphy()->getData($data['query']);

		$data = $data ? $data : false;

		$args = array('giphies' => $data);

		$contents = CK::getTemplateContents('giphy.browser', $args);

		$output = new stdClass();
		$output->contents = $contents;

		return $this->output($output);
	}

	/**
	 * Creates a new conversation
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function create()
	{
		ES::checkToken();
		ES::requireLogin();

		$userId = $this->my->id;

		$data = $this->input->getArray('post');
		$data['uid'] = $data['recipientId'];

		$message = $data['message'];

		$data = $this->lib->create($userId, $data);

		// skip the process if detected there got any error
		if (isset($data->code) && $data->code == '403') {
			return $this->showError($data);
		}

		$data->messages = array($data->message);

		// Format the message for CK
		$data->messages = $this->formatMessages($data, $message);

		$args = array('data' => $data);

		$contents = CK::getTemplateContents('message', $args);

		$output = new stdClass();
		$output->id = $data->id;
		$output->contents = $contents;

		return $this->output($output);
	}

	/**
	 * Responds in an existing conversation
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function reply()
	{
		ES::checkToken();
		ES::requireLogin();

		$date = JFactory::getDate();

		$id = $this->input->get('uid', 0, 'int');
		$userId = $this->my->id;
		$data = $this->input->getArray('post');
		$message = $this->input->getString('message');

		// determine whether there got any file attached
		$attachments = isset($data['upload-id']) ? $data['upload-id'] : null;

		if (empty($message) && !$attachments) {
			$this->showError(JText::_('PLG_CK_ADD_MESSAGE'));
		}

		if (!$id) {
			$this->showError(JText::_('PLG_CK_INVALID_ID'));
		}

		$limit = $this->input->get('limit', $this->params->get('conversation_limit'), 'default');
		$start = $this->input->get('limitstart', 0, 'int');

		$options = array(
			'limit' => $limit,
			'start' => $start
		);

		// check if user allow to reply to this conversation or not.
		$conversation = ES::conversation($id);
		$allowed = $conversation->canCreate();

		if (!$allowed) {

			$obj = new stdClass();

			$obj->code = '403';
			$obj->message = JText::_('PLG_CK_NOT_ALLOWED_TO_SEND_MESSAGE');

			$errMessage = $conversation->getError();
			if ($errMessage) {
				$obj->message = JText::_($errMessage);
			}

			$output = $this->output($obj);
			return $output;
		}

		// here we need to unset the uid from data as
		// ES conversation treat uid as recipient id.
		if (isset($data['uid'])) {
			unset($data['uid']);
		}

		$data = $this->lib->reply($userId, $id, $data);

		// Format the message for CK
		$data->messages = $this->formatMessages($data, $message);

		if (isset($data->code) && $data->code == '403') {
			return $this->output($data);
		}

		$args = array('data' => $data);

		$output = new stdClass();
		$output->id = $id;
		$output->contents = CK::getTemplateContents('message', $args);

		return $this->output($output);
	}

	/**
	 * Format the messages after create or reply
	 *
	 * @since	2.0
	 * @access	public
	 */
	public function formatMessages($data, $message)
	{
		$messages = array();

		if ($data->messages) {
			foreach ($data->messages as $item) {
				// Need to get the giphy preview template from CK instead of ES to work properly
				if ($item->type == 'giphy') {
					$item->text = CK::getTemplateContents('giphy.preview', array('giphy' => $message));
				}

				$messages[] = $item;
			}
		}

		return $messages;
	}

	/**
	 * Allows caller to update the title of the conversation
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function updateTitle()
	{
		ES::checkToken();
		ES::requireLogin();

		$conversationId = $this->input->get('conversationId', 0, 'int');
		$title = $this->input->get('title', '', 'string');

		$this->lib->updateTitle($conversationId, $this->my->id, $title);


		$result = new stdClass();
		$result->title = $this->getTitle($title);

		return $this->output($result);
	}

	/**
	 * Outputs json data
	 *
	 * @since	1.1.0
	 * @access	public
	 */
	public function output($data)
	{
		header('Content-type: application/json; UTF-8');
		echo json_encode($data);
		exit;
	}

	/**
	 * Renders a json response when the system needs to trigger an error
	 *
	 * @since	1.0.4
	 * @access	public
	 */
	public function showError($string)
	{
		$data = new stdClass();
		$data->code = '500';
		$data->message = $string;
		$output = $this->output($data);
		return $output;
	}
}
